/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AddressbooksSynchronizer.
 *
 * The Initial Developer of the Original Code is Gnter Gersdorf.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Gnter Gersdorf <G.Gersdorf@ggbs.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
var gAccountManager = Components.classes["@mozilla.org/messenger/account-manager;1"]
    .getService(Components.interfaces.nsIMsgAccountManager);
var sMsgComposeService = Components.classes["@mozilla.org/messengercompose;1"]
    .getService(Components.interfaces.nsIMsgComposeService);
var nsIMsgCompDeliverMode = Components.interfaces.nsIMsgCompDeliverMode;

var gBookCount;
var gMabFiles;
var gImapFolder;
var gImapFolderURI;
var gAccount;
var gIdentity;
var gSavedDraftFolder;
var gMailSession;
var gMsgCompose;
var gMessenger;

var QPEncoder;

// from mailnews\base\public\nsMsgMessageFlags.h messenger/mailWindowOverlay.js
//#define MSG_FLAG_READ     0x0001    /* has been read */
const abs_MSG_FLAG_READ              = 0x000001;
//  already defined in mailWindowOverlay.js

function imapDone() {
  gImapFolder.compact(null, null);  // only needed on upload
}

///////////////////////////////////////////////////////////////////////
//  Save mab files to imap folder


function imapSave() {
  if (!checkOffline()) return;
  var imapFolderuri=pref.getCharPref( "extensions.addressbookssync.imapfolder" );
  gImapFolder = GetMsgFolderFromUri(imapFolderuri, true); //nsIRDFResource, nsIMsgFolder
  if (!gImapFolder) {
    statustxt(errorstr(-3), true, true);
    killMe();
    return;
  }
  QPEncoder = new QuotedPrintableEncoder();

  var server=gImapFolder.server;
  gAccount=gAccountManager.FindAccountForServer(server);
  gIdentity=gAccount.defaultIdentity;

  gTempName='absmail.eml';
  cleanUp=imapDone; // register cleanup function

  gMabFiles=syncedAddressbooks();
  imapSaveNextBook();

  return;
}


function imapSaveNextBook() {
  var mabFile=gMabFiles.pop();
  if (!mabFile || gAbort) {
    killMe(); // kill status popup
    return;
  }

  var cur=getAddressBook(mabFile);
  var mabName=cur.directoryProperties.description;
  var uri = cur.directoryProperties.URI;

  var useMabName;
  if (uri!=kPersonalAddressbookURI && uri!=kCollectedAddressbookURI)
    useMabName = mabName+'.mab';
  else
    useMabName = '_tb_'+cur.directoryProperties.fileName;

  statustxt('upload '+mabName+'.', true, false);  // 1. dot: started

//  get old messages for this addressbook
  var deleteMsgs=Components.classes["@mozilla.org/supports-array;1"]
    .createInstance(Components.interfaces.nsISupportsArray);
  var msgId=getMsgId(useMabName);
  var msgs=gImapFolder.getMessages(null);  //simpleEnumerator
  while (msgs.hasMoreElements()) {
    var msg=msgs.getNext();
    msg.QueryInterface(Components.interfaces.nsIMsgDBHdr);
    if (msg.messageId == msgId) deleteMsgs.AppendElement(msg);
  }


  // Create a message and save in IMAP Folder

    // One possibility to do this is to use nsIMsgComposeService.InitCompose()
    //  with nsIMsgComposeParams.composeFields() to create the Message
    //  and use nsIMsgCompose.SendMsg() to 'send' the message.
    //  This works quite well but has some major drawbacks:
    //    - The message can only put in the drafts or template folders
    //      and has to be moved afterwards (or, as i implemented it, temporarily
    //      change the drafts folder)
    //    - If the account is set to sign and/or encrypt the message, this message will
    //      be signed and/or encrypted. Probably, other settings (or extensions like
    //      Enigmail will also alter the message. This makes it difficult or even nearly
    //      impossible to retrieve the attachment on download
    // The other possibility is to create a message 'by hand' and store it
    //  in a temporary file. With nsIMsgFolder.copyFileMessage(), the tempfile
    //  could be load and stored in the correct IMAP-Folder
    //  This is whats implemented here

  var qpName=QPEncoder.Encode(useMabName);  // quoted-printable encode the filename
  var time=Date.now();

  var header='';
  header+='Message-ID: '+getMsgId(useMabName)+'\r\n';
  header+='Date: '+Date(time)+'\r\n';
  header+='From: "Addressbooks Synchronizer"\r\n';
  header+='To: "Addressbooks Synchronizer"\r\n';
  header+='Subject: Addressbooks Synchronizer: =?ISO-8859-1?Q?'+qpName+'?=\r\n';
  header+='Content-Type: multipart/mixed; boundary="------------'+time+'"\r\n';
  header+='\r\n';
  header+='This is a multi-part message in MIME format.\r\n';
  header+='--------------'+time+'\r\n';
  header+='Content-Type: text/plain; charset=ISO-8859-1\r\n';
  header+='Content-Transfer-Encoding: quoted-printable\r\n';
  header+='\r\n';
  header+='Addressbooks Synchronizer: '+qpName+'\r\n';
  header+='Do not delete this message!!!\r\n';
  header+='\r\n';
  header+='--------------'+time+'\r\n';
  header+='Content-Type: text/plain; name="'+qpName+'"\r\n';
  header+='Content-Transfer-Encoding: 7bit\r\n';
  header+='Content-Disposition: inline; filename="'+qpName+'"\r\n';
  header+='\r\n';

  var tailer='';
  tailer+='\r\n';
  tailer+='--------------'+time+'\r\n';

  var nsIMabFile=getDir('ProfD');
  nsIMabFile.appendRelativePath( mabFile );
  abData=readMabData(nsIMabFile);
  if (!abData) {
    statustxt(strings.getString('notcopied'), false, false);
    imapSaveNextBook();
    return;
  }
  statustxt(".", false, false); // 2. dot: mab data loaded

  var nsITmpFile=getDir('TmpD');
  nsITmpFile.appendRelativePath(gTempName);
  if (!writeMabData(nsITmpFile, header+abData+tailer)) {
    statustxt(strings.getString('notcopied'), false, false);
    imapSaveNextBook();
    return;
  }
  statustxt(".", false, false); // 3. dot: tmp file written

  var fileSpec = Components.classes["@mozilla.org/filespec;1"].createInstance();
  fileSpec = fileSpec.QueryInterface( Components.interfaces.nsIFileSpec);
  fileSpec.nativePath = nsITmpFile.path;

  // listen for the end of a 'store to imap folder'
  var copyServiceListener = {
      deleteMsgs: null,
      QueryInterface: function(iid) {
dump('ABS: IMAP: QueryInterface called\n');
        if (iid.equals(Components.interfaces.nsIMsgCopyServiceListener) ||
            iid.equals(Components.interfaces.nsISupports))
          return this;
        throw Components.results.NS_NOINTERFACE;
        return 0;
      },

      OnProgress:  function(progress, progressMax) {
dump('ABS: IMAP: OnProgress called: '+progress+' '+progressMax+'\n');
        // never called!!
      },

      OnStartCopy: function() {
dump('ABS: IMAP: OnStartCopy called\n');
        // never called!!
      },

      OnStopCopy:  function(status) {
dump('ABS: IMAP: OnStopCopy called: status='+status+'\n');
        // never called!!
        if (!status) {
          statustxt(".", false, false); // 5. dot: copied successfully
          // call to SetMessageKey follows
        } else {
//          statustxt(strings.getString('notcopied'), false, true);
          statustxt(errorstr(status), false, true);
          setTimeout(imapSaveNextBook, 100);
          //Failed
        }
      },

      SetMessageKey:  function(key) {
dump('ABS: IMAP: SetMessageKey called: key='+key+'\n');
        // store is finished, delete old messages for this mab and start store of next mab
        statustxt(strings.getString('copied'), false, false);
        if (this.deleteMsgs.Count()>0)
          gImapFolder.deleteMessages ( this.deleteMsgs, null, true, false, null, false);
        // this timegap is really needed!!
        setTimeout(imapSaveNextBook, 100);
      }
  }
  copyServiceListener.deleteMsgs=deleteMsgs;
  gImapFolder.copyFileMessage(fileSpec, null, false, abs_MSG_FLAG_READ, null, copyServiceListener);

  statustxt(".", false, false); // 4. dot: copy to IMAP folder started

  // next mabfile is started, when listener is informed about the end of the operation
  return;
}

/////////////////////////////////////////////////////////////////////////
//  Load mab files from imap folder

// from messageWindow@48
// wait till all headers are read for the imap folder
var syncListener = {
  OnItemAdded: function(parentItem, item) {},
  OnItemRemoved: function(parentItem, item) {},
  OnItemPropertyChanged: function(item, property, oldValue, newValue) {},
  OnItemIntPropertyChanged: function(item, property, oldValue, newValue) {},
  OnItemBoolPropertyChanged: function(item, property, oldValue, newValue) {},
  OnItemUnicharPropertyChanged: function(item, property, oldValue, newValue){},
  OnItemPropertyFlagChanged: function(item, property, oldFlag, newFlag) {},
  OnItemEvent: function(folder, event) {
     var eventType = event.toString();

    if (eventType == "FolderLoaded") {
      if (folder) {
        var uri = folder.URI;
        if (uri == gImapFolderURI) {
          gMailSession.RemoveFolderListener(this);
          var msgFolder = folder.QueryInterface(Components.interfaces.nsIMsgFolder);
          if (msgFolder) {
            msgFolder.endFolderLoading();
            imapLoadStage2();
          }
        }
      }
    }
  }
}

function imapLoad() {
  if (!checkOffline()) return;
  gImapFolderURI=pref.getCharPref( "extensions.addressbookssync.imapfolder" );
  gImapFolder = GetMsgFolderFromUri(gImapFolderURI, true);
  if (!gImapFolder || gImapFolder.server.type!='imap') {
    statustxt(errorstr(-3), true, true);
    killMe();
    return;
  }

  gTempName='abstemp.mab';
  cleanUp=imapDone; // register cleanup function

  var server=gImapFolder.server;
  gAccount=gAccountManager.FindAccountForServer(server);
  gIdentity=gAccount.defaultIdentity;

  statustxt(strings.getString('loadimap')+'..',true,false);

  // first be sure all header are available
  // when they are, the syncListener calls starts the next stage
  var nsIFolderListener = Components.interfaces.nsIFolderListener;
  gMailSession = Components.classes["@mozilla.org/messenger/services/session;1"].
    getService(Components.interfaces.nsIMsgMailSession);
  gMailSession.AddFolderListener(syncListener, nsIFolderListener.event);
  gImapFolder.startFolderLoading();
  gImapFolder.updateFolder(null);

  return;
}

function imapLoadStage2() {
  statustxt(strings.getString('ok'),false,false);
  if (gAbort) { killMe(); return; }

  gMessenger = Components.classes['@mozilla.org/messenger;1'].createInstance().
    QueryInterface(Components.interfaces.nsIMessenger);

  gMabFiles=syncedAddressbooks();
  imapLoadNextBook();
  return;
}

function imapLoadNextBook() {
  if (gAbort) { killMe(); return; }
  var mabFile=gMabFiles.pop();
  if (!mabFile) {
    killMe(); // kill status popup
    return;
  }

  var cur=getAddressBook(mabFile);
  var mabName=cur.directoryProperties.description;
  var uri = cur.directoryProperties.URI;

  var useMabName;
  if (uri!=kPersonalAddressbookURI && uri!=kCollectedAddressbookURI)
    useMabName = mabName+'.mab';
  else
    useMabName = '_tb_'+cur.directoryProperties.fileName;

  statustxt(mabName+'.', true, false);  // 1. dot: start

  // find the correct message (looking for the messageId)
  var msgId = getMsgId(useMabName);
  var msgs=gImapFolder.getMessages(null);  //simpleEnumerator
  var theMsg=null;
  var theMsgKey=0;
  while (msgs.hasMoreElements()) {
    var msg=msgs.getNext();
    msg.QueryInterface(Components.interfaces.nsIMsgDBHdr);
    // msg.messageKey are the id's as used by the imapserver
    // find the last corresponding message (but there should only be one)
    if (msg.messageId == msgId && msg.messageKey>theMsgKey) {
      theMsg=msg;
      theMsgKey=msg.messageKey;
    }
  }
  if (!theMsg) {
    statustxt(strings.getString('nomessage'), false, false);
    imapLoadNextBook();
  }
  statustxt('.', false, false); // 2. dot: message found

  var msgURI=gImapFolder.getUriForMsg(theMsg);
  var ms=gMessenger.messageServiceFromURI(msgURI);

  // Extract attachment from the message and use them to recreate addressbooks

    // One possibility to do this is to use of nsIMessenger.saveAttachmentToFolder
    //    But for this we have to use a url to the attachment which encodes the part
    //    number of the attachment. But there is no way to get this number (these
    //    informations are only available for messages which are shown in a message
    //    window). If we have used the first possibility for crwating the message
    //    (see comment in imapSaveNextBook()), the part number may vary depending
    //    on signing, encryption etc. With the current implementation, the partnumber
    //    is always '1.2'. But using saveAttachmentToFolder has another drawback:
    //    it stores the data asynchronously and i found no way to be informed about
    //    the store being finished.
    //  The other possibility and now implemented is the use of
    //    nsIMsgMessageService.streamMessage(). This gives me the whole message in
    //    a string. But i know the structure of the message and can extract the
    //    addressbook.


  var streamListener = {
    body: '',
    mabName: null,
    mabFile: null,
    onDataAvailable: function(request, context, inputStream, offset, count){
//statustxt(' onDataAvailable-'+offset+'-'+count+' ', false, false);
      try {
        var sis=Components.classes["@mozilla.org/scriptableinputstream;1"].
          createInstance(Components.interfaces.nsIScriptableInputStream);
        sis.init(inputStream);
        this.body+=sis.read(count);
      } catch(ex) { alert(ex+'\noffset='+offset+' count='+count); }
    },
    onStartRequest: function(request, context) {
//statustxt(' onStartRequest ', false, false);
      this.body='';
    },
    onStopRequest: function(aRequest, aContext, aStatusCode) {
//statustxt(' onStopRequest-'+aRequest.status+'-'+aStatusCode+' ', false, false);
      if (!aStatusCode) {
        statustxt('.', false, false); // 3. dot: message read

        // extract mab data from message
        var boundary=this.body.match(/boundary="(.*?)"/);
        boundary='--'+boundary[1];
        var ind=this.body.indexOf(boundary);     // start of text part
        ind=this.body.indexOf(boundary, ind+1);  // start of mab part
        ind=this.body.indexOf('//', ind+1);      // mab starts with '//'
        var lind=this.body.indexOf(boundary, ind+1);  // end of mab part
        this.body=this.body.substring(ind,lind);

        // write mab to temporary file
        var nsITmpFile=getDir('TmpD');
        nsITmpFile.appendRelativePath(gTempName);
        writeMabData(nsITmpFile, this.body);
        statustxt('.', false, false); // 4. dot: temp file created

        // replace addressbook data
        var ret;
        if ((ret=replaceAddressBook(this.mabName, this.mabFile, nsITmpFile)))
          statustxt(ret, false, true);
        else
          statustxt(strings.getString('copied'), false, false);
      } else {
        statustxt(strings.getString('notcopied'), false, true);
      }
      imapLoadNextBook();
    }
  };

  streamListener.mabName=mabName;
  streamListener.mabFile=mabFile;
	var aurl = new Object();
	ms.CopyMessage(msgURI, streamListener, false, null, null, aurl);
//alert(aurl.value.spec);
  return;
}

/////////////////////////////////////////////////////////////////////////

function getMsgId(theName) {
  return encodeURI(theName+'@AddressBooksSynchronizer');
}

/*
 * from http://www.snippetcenter.org/de/conversion-to-content-transfer-encoding-quoted-printable-s1786.aspx
 */
function Encode() {
  var QPEncoder = new QuotedPrintableEncoder();
  Output.value = QPEncoder.Encode(Input.value);
}
function QuotedPrintableEncoder() {
  this.MaxLineSize = 76;
  this.Encode = QuotedPrintableEncode;
  this.EncodeLine = QuotedPrintableEncodeLine;
  this.EncodeCharacter = QuotedPrintableEncodeCharacter;
  // Exclamation point through less than, and greater than through tilde
  // Tab is not listed here as UnAltered but could be
  this.UnAltered = String.fromCharCode(32) + String.fromCharCode(60) + String.fromCharCode(62) + String.fromCharCode(126);
  this.UnAlteredEnd = String.fromCharCode(33) + String.fromCharCode(60) + String.fromCharCode(62) + String.fromCharCode(126);
}
function QuotedPrintableEncode(InputText) {
  var re = new RegExp("\r\n","g");
  var SplitText, i, s="";
  SplitText = InputText.split(re);
  for (i=0; i<SplitText.length; ++i) {
    s += this.EncodeLine(SplitText[i]);
    if (i<SplitText.length-1) s+="\r\n";
  }
  return s;
}
function QuotedPrintableEncodeLine(Text) {
  var SplitText, i, s="", c=256;
  if (Text.length == 0) return "";
  for (i=0; i<Text.length-1; ++i)
    s += this.EncodeCharacter(Text.charCodeAt(i), this.UnAltered);
  // Encode last character; if space, encode it
  s += this.EncodeCharacter(Text.charCodeAt(Text.length-1), this.UnAlteredEnd);
  if (s.length <= this.MaxLineSize) return s; // Split into lines of MaxLineSize characters or less
  SplitText = s.slice(0, this.MaxLineSize);
  i = this.MaxLineSize;
  while (i<s.length) {
    SplitText += "=\r\n"+ s.slice(i, i+this.MaxLineSize);
    i += this.MaxLineSize;
  }
  return SplitText;
}
function QuotedPrintableEncodeCharacter(Character, UnAltered) {
  var i, x, Alter=true;
  for (i=0; i<UnAltered.length; i+=2)
    if (Character >= UnAltered.charCodeAt(i) && Character <= UnAltered.charCodeAt(i+1))
      Alter=false;
  if (!Alter) return String.fromCharCode(Character);
  x = Character.toString(16).toUpperCase();
  return (x.length == 1) ? "=0" + x : "=" + x;
}
